// BlockingQueue.java
//
// CompSci 141 / CSE 141 / Informatics 101 Spring 2013
// Code Example
//
// This is an implementation of a "blocking queue."  Blocking queues are
// different from typical ones in the sense that they're intended to be
// used in concurrent contexts, meaning two things:
//
// (1) Synchronization becomes more important, since we want to disallow
//     multiple threads from simultaneously accessing it in ways that can
//     cause problems.  Here, we've synchronized aggressively, making sure
//     that no two threads can ever be in a BlockingQueue method at the
//     same time.  A better, though somewhat more difficult, approach is
//     to lock more carefully so that, for example, one thread can be
//     enqueuing at one end of a linked list while another can be dequeuing
//     from the other end.  And the biggest challenge of all is implementing
//     a "lock-free" data structure -- one that uses the guarantees built into
//     Java's underlying memory model to ensure correct behavior even in the
//     absence of locking; this is no small task, and there aren't many folks
//     who can implement that kind of data structure correctly (but, luckily,
//     the Java library contains a few of them, so us mere mortals don't have
//     to!).
//
// (2) Attempting to dequeue when the queue is empty is not necessarily
//     a problem.  Rather than throwing an exception in this case, another
//     possible solution is to instead have the calling thread block until
//     some other thread enqueues an element; this is the approach shown
//     below.  Most typically, a blocking queue implementation provides
//     both options, as well as other possible ones, such as blocking for
//     a certain amount of time before giving up.
//
// In practice, we're much better off using a pre-built implementation of
// something like this when we can.  Luckily, the Java library has
// wonderful implementations of these kinds of things in its
// java.util.concurrent package (e.g., LinkedBlockingQueue), so we should
// tend to want those whenever we can use them.  But for illustrative
// purposes, consider this implementation of a blocking queue.
//
// When you might use a blocking queue is when one or more threads need to
// perform a sequence of tasks, one after another, where they'll sometimes
// have work to do and sometimes not.  The queue is a mechanism to allow the
// thread to have work "queued up" if the tasks come in faster than the thread
// can handle them for a while; the *blocking* aspect of the queue will allow
// the thread to lie dormant when it has no work to do.  Queues show up in
// many places in distributed or concurrent systems for exactly this reason.

import java.util.LinkedList;


public class BlockingQueue<E>
{
	// A LinkedList containing the items that are currently in the queue.
	// It's important to note that the LinkedList class in Java is not
	// "thread safe," meaning that it's not safe for multiple threads to
	// simultaneously access it.  This won't be a problem for us, since
	// we've synchronized all of the methods that might access it; no
	// two threads will ever be able to be in a BlockingQueue method,
	// so no two threads will ever be able to access the LinkedList at
	// the same time.
	private LinkedList<E> items;
	

	public BlockingQueue()
	{
		items = new LinkedList<E>();
	}
	

	// enqueue() enqueues an element into the queue.
	public synchronized void enqueue(E e)
	{
		// Enqueue an item into the queue by adding it to the end of the
		// list of items.
		items.add(e);
		
		// We call notify() here so that if there are other threads
		// currently waiting inside of dequeue(), one of them will
		// awaken and be eligible to dequeue this newly-added item.
		// If no threads are waiting, this notification will simply
		// go unheard, with no harm done.
		//
		// Note that you have to synchronize on an object before you
		// call notify() on it.  This is no problem here; since we're
		// in a synchronized method, we've synchronized on "this"
		// -- the same object we're calling notify() on -- already.
		notify();
	}
	

	// dequeue() dequeues an element from the queue, blocking indefinitely
	// if there are no elements available.  We also support interruption,
	// so that this method will throw an InterruptedException if the
	// calling thread is interrupted while waiting for an element.
	public synchronized E dequeue()
	throws InterruptedException
	{
		// It looks a little strange for us to have used a "while" loop
		// here instead of an "if" statement.  The reason is that the
		// wait() method is subject to what are called "spurious wakeups".
		// It's possible for wait() to wake up from time to time even if
		// notify() was never called.  So we ensure that the size really
		// has changed each time we wake up, going back to waiting if
		// we woke up by accident.
		while (items.size() == 0)
		{
			wait();
		}

		// Once we get here, we know there was either an item in the queue
		// to begin with, or that we waited and one eventually showed up.
		// Because this method (and the others) are synchronized, there
		// is no chance that the only element will be "stolen" by another
		// thread beteween the time this thread wakes up from its wait()
		// and the time it gets here.
		return items.removeFirst();
	}


	// size() returns the number of elements in this queue.  Even this is
	// potentially problematic unless we synchronize it, since it's not
	// safe to call size() while another thread is modifying the linked
	// list.
	public synchronized int size()
	{
		return items.size();
	}
}
